@@ -86,7 +86,7 @@ def _read_ssh_ed25519_pubkey(keydata):
         ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIJukDU5fqXv/yVhSirsDWsUFyOodZyCSLxyitPPzWJW9 oberstet@office-corei7
 
     :param keydata: The OpenSSH Ed25519 public key data to parse.
-    :type keydata: unicode
+    :type keydata: str
 
     :returns: pair of raw public key (32 bytes) and comment
     :rtype: tuple
@@ -159,7 +159,7 @@ def _read_ssh_ed25519_privkey(keydata):
 
 
     :param keydata: The OpenSSH Ed25519 private key data to parse.
-    :type keydata: unicode
+    :type keydata: str
 
     :returns: pair of raw private key (32 bytes) and comment
     :rtype: tuple
@@ -352,6 +352,7 @@ def _verify_signify_ed25519_signature(pubkey_file, signature_file, message):
 
 if HAS_CRYPTOSIGN:
 
+    @util.public
     class SigningKey(object):
         """
         A cryptosign private key for signing, and hence usable for authentication or a
@@ -375,8 +376,10 @@ if HAS_CRYPTOSIGN:
             self._can_sign = isinstance(key, signing.SigningKey)
 
         def __str__(self):
-            return u'Key(can_sign={}, comment="{}", public_key={})'.format(self.can_sign(), self.comment(), self.public_key())
+            comment = u'"{}"'.format(self.comment()) if self.comment() else None
+            return u'Key(can_sign={}, comment={}, public_key={})'.format(self.can_sign(), comment, self.public_key())
 
+        @util.public
         def can_sign(self):
             """
             Check if the key can be used to sign.
@@ -386,21 +389,23 @@ if HAS_CRYPTOSIGN:
             """
             return self._can_sign
 
+        @util.public
         def comment(self):
             """
             Get the key comment (if any).
 
             :returns: The comment (if any) from the key.
-            :rtype: unicode or None
+            :rtype: str or None
             """
             return self._comment
 
+        @util.public
         def public_key(self, binary=False):
             """
             Returns the public key part of a signing key or the (public) verification key.
 
             :returns: The public key in Hex encoding.
-            :rtype: unicode or None
+            :rtype: str or None
             """
             if isinstance(self._key, signing.SigningKey):
                 key = self._key.verify_key
@@ -412,6 +417,7 @@ if HAS_CRYPTOSIGN:
             else:
                 return key.encode(encoder=encoding.HexEncoder).decode('ascii')
 
+        @util.public
         def sign(self, data):
             """
             Sign some data.
@@ -437,6 +443,7 @@ if HAS_CRYPTOSIGN:
             # the signature
             return txaio.create_future_success(sig.signature)
 
+        @util.public
         def sign_challenge(self, session, challenge):
             """
             Sign WAMP-cryptosign challenge.
@@ -445,7 +452,7 @@ if HAS_CRYPTOSIGN:
             :type challenge: instance of autobahn.wamp.types.Challenge
 
             :returns: A Deferred/Future that resolves to the computed signature.
-            :rtype: unicode
+            :rtype: str
             """
             if not isinstance(challenge, Challenge):
                 raise Exception("challenge must be instance of autobahn.wamp.types.Challenge, not {}".format(type(challenge)))
@@ -488,6 +495,7 @@ if HAS_CRYPTOSIGN:
 
             return d2
 
+        @util.public
         @classmethod
         def from_raw_key(cls, filename, comment=None):
             """
@@ -501,9 +509,9 @@ if HAS_CRYPTOSIGN:
                 dd if=/dev/urandom of=client02.key bs=1 count=32
 
             :param filename: Filename of the key.
-            :type filename: unicode
+            :type filename: str
             :param comment: Comment for key (optional).
-            :type comment: unicode or None
+            :type comment: str or None
             """
             if not (comment is None or type(comment) == six.text_type):
                 raise Exception("invalid type {} for comment".format(type(comment)))
@@ -520,6 +528,7 @@ if HAS_CRYPTOSIGN:
             key = signing.SigningKey(keydata)
             return cls(key, comment)
 
+        @util.public
         @classmethod
         def from_ssh_key(cls, filename):
             """
